from pathlib import Path
from unittest.mock import patch

from airflow_balancer import BalancerConfiguration
from airflow_balancer.testing import pools
from airflow_balancer.ui.functions import get_hosts_from_yaml, get_yaml_files
from airflow_balancer.ui.standalone import build_app, main


class TestAirflowPlugin:
    def test_plugin(self):
        from airflow_balancer.ui.viewer import AirflowBalancerViewerPlugin, AirflowBalancerViewerPluginView

        AirflowBalancerViewerPluginView()
        AirflowBalancerViewerPlugin()

    # def test_plugin_view(self):
    #     with patch("airflow_balancer.ui.viewer.expose") as mock_expose, \
    #         patch("airflow_balancer.ui.viewer.has_access") as mock_has_access, \
    #             patch("airflow_balancer.ui.viewer.request") as mock_request:
    #         from airflow_balancer.ui.viewer import AirflowBalancerViewerPluginView
    #         mock_expose.side_effect = lambda x: lambda f: f
    #         mock_has_access.side_effect = lambda x: lambda f: f
    #         pv = AirflowBalancerViewerPluginView()
    #         pv.home()


class TestPluginFunctions:
    def test_plugin_functions_get_yamls(self):
        root = Path(__file__).parent
        assert get_yaml_files(root / "config") == [
            Path(root) / "config/extensions/default.yaml",
            Path(root) / "config/extensions/balancer.yaml",
            Path(root) / "config/extensions/second.yaml",
        ]

    def test_plugin_functions_load_yamls(self):
        root = Path(__file__).parent
        with pools():
            assert isinstance(BalancerConfiguration.load(Path(root) / "config/extensions/default.yaml"), BalancerConfiguration)
            assert isinstance(BalancerConfiguration.load(Path(root) / "config/extensions/balancer.yaml"), BalancerConfiguration)

    def test_plugin_functions_get_hosts(self):
        root = Path(__file__).parent
        assert (
            get_hosts_from_yaml(Path(root) / "config/extensions/default.yaml")
            == '{"hosts":[{"name":"host0","username":"timkpaine","password":null,"password_variable":null,"password_variable_key":null,"key_file":"/home/airflow/.ssh/id_rsa","os":"ubuntu","pool":"host0","size":8,"queues":["workers"],"tags":[]},{"name":"host1","username":"timkpaine","password":null,"password_variable":null,"password_variable_key":null,"key_file":"/home/airflow/.ssh/id_rsa","os":"ubuntu","pool":"host1","size":16,"queues":["primary"],"tags":[]},{"name":"host2","username":"timkpaine","password":null,"password_variable":null,"password_variable_key":null,"key_file":"/home/airflow/.ssh/id_rsa","os":"ubuntu","pool":"host2","size":16,"queues":["workers"],"tags":[]},{"name":"host3","username":"timkpaine","password":null,"password_variable":null,"password_variable_key":null,"key_file":"/home/airflow/.ssh/id_rsa","os":"macos","pool":"host3","size":8,"queues":["workers"],"tags":[]}],"ports":[{"name":"","host":{"name":"host1","username":"timkpaine","password":null,"password_variable":null,"password_variable_key":null,"key_file":"/home/airflow/.ssh/id_rsa","os":"ubuntu","pool":"host1","size":16,"queues":["primary"],"tags":[]},"host_name":"host1","port":8000,"tags":[]},{"name":"named-port","host":{"name":"host1","username":null,"password":null,"password_variable":null,"password_variable_key":null,"key_file":null,"os":"ubuntu","pool":null,"size":16,"queues":["primary"],"tags":[]},"host_name":"","port":8080,"tags":[]},{"name":"","host":{"name":"host2","username":"timkpaine","password":null,"password_variable":null,"password_variable_key":null,"key_file":"/home/airflow/.ssh/id_rsa","os":"ubuntu","pool":"host2","size":16,"queues":["workers"],"tags":[]},"host_name":"host2","port":8793,"tags":[]}],"default_username":"timkpaine","default_password":null,"default_password_variable":null,"default_password_variable_key":null,"default_key_file":"/home/airflow/.ssh/id_rsa","primary_queue":"primary","secondary_queue":"workers","default_queue":"default","default_size":8,"override_pool_size":false,"create_connection":false}'
        )
        assert (
            get_hosts_from_yaml(Path(root) / "config/extensions/balancer.yaml")
            == '{"hosts":[{"name":"host0","username":"test","password":null,"password_variable":null,"password_variable_key":null,"key_file":"/home/airflow/.ssh/id_rsa","os":"ubuntu","pool":"host0","size":8,"queues":["workers"],"tags":[]},{"name":"host1","username":"test","password":null,"password_variable":null,"password_variable_key":null,"key_file":"/home/airflow/.ssh/id_rsa","os":"ubuntu","pool":"host1","size":16,"queues":["primary"],"tags":[]},{"name":"host2","username":"test","password":null,"password_variable":null,"password_variable_key":null,"key_file":"/home/airflow/.ssh/id_rsa","os":"ubuntu","pool":"host2","size":16,"queues":["workers"],"tags":[]},{"name":"host3","username":"test","password":null,"password_variable":null,"password_variable_key":null,"key_file":"/home/airflow/.ssh/id_rsa","os":"macos","pool":"host3","size":8,"queues":["workers"],"tags":[]}],"ports":[{"name":"","host":{"name":"host1","username":"test","password":null,"password_variable":null,"password_variable_key":null,"key_file":"/home/airflow/.ssh/id_rsa","os":"ubuntu","pool":"host1","size":16,"queues":["primary"],"tags":[]},"host_name":"host1","port":8000,"tags":[]},{"name":"named-port","host":{"name":"host1","username":null,"password":null,"password_variable":null,"password_variable_key":null,"key_file":null,"os":"ubuntu","pool":null,"size":16,"queues":["primary"],"tags":[]},"host_name":"","port":8080,"tags":[]},{"name":"","host":{"name":"host2","username":"test","password":null,"password_variable":null,"password_variable_key":null,"key_file":"/home/airflow/.ssh/id_rsa","os":"ubuntu","pool":"host2","size":16,"queues":["workers"],"tags":[]},"host_name":"host2","port":8793,"tags":[]}],"default_username":"test","default_password":null,"default_password_variable":null,"default_password_variable_key":null,"default_key_file":"/home/airflow/.ssh/id_rsa","primary_queue":"primary","secondary_queue":"workers","default_queue":"default","default_size":8,"override_pool_size":false,"create_connection":false}'
        )


class TestStandaloneUI:
    def test_standalone_ui(self):
        # Test the build_app function
        app = build_app()
        assert app is not None

    def test_launch(self):
        # Test the main function
        with patch("airflow_balancer.ui.standalone.run") as mock_run:
            main()
            mock_run.assert_called_once()

    def test_main(self):
        import airflow_balancer.ui.standalone.__main__  # noqa: F401



from __future__ import annotations

from pathlib import Path

from hydra.errors import InstantiationException

from airflow_balancer import BalancerConfiguration
from airflow_balancer.testing import pools

__all__ = (
    "get_hosts_from_yaml",
    "get_yaml_files",
)


def get_hosts_from_yaml(yaml: str) -> list[str]:
    # Process the yaml
    yaml_file = Path(yaml).resolve()
    try:
        inst = BalancerConfiguration.load(yaml_file)
    except InstantiationException:
        # Mock SQL connections to instantiate
        with pools():
            inst = BalancerConfiguration.load(yaml_file)
    for host in inst.hosts:
        if host.password:
            host.password = "***"
    if inst.default_password:
        inst.default_password = "***"
    for port in inst.ports:
        if port.host.password:
            port.host.password = "***"
    return str(inst.model_dump_json(serialize_as_any=True))


def get_yaml_files(dags_folder: str) -> list[Path]:
    # Look for yamls inside the dags folder
    yamls = []
    base_path = Path(dags_folder)

    # Look if the file directly instantiates a BalancerConfiguration
    for path in base_path.glob("**/*.yaml"):
        if path.is_file():
            if "_target_: airflow_balancer.BalancerConfiguration" in path.read_text():
                yamls.append(path)
    len_yamls = len(yamls)
    len_yamls_last = 0
    # If we have yamls, look for any that reference them
    while len_yamls != len_yamls_last:
        for path in base_path.glob("**/*.yaml"):
            if path.is_file() and path not in yamls:
                # Check and see if this references any existing yamls
                for yaml in yamls:
                    if path.parent == yaml.parent and f"{yaml.stem}@" in path.read_text():
                        yamls.append(path)
                        break
        len_yamls_last = len_yamls
        len_yamls = len(yamls)
    return yamls



from typing import List, Optional

from pydantic import BaseModel, Field, model_validator
from typing_extensions import Self

from .host import Host

__all__ = ("Port",)


class Port(BaseModel):
    name: str = ""
    host: Optional[Host] = None
    host_name: str = ""
    port: int = Field(default=None, ge=1, le=65535, description="Port number")
    tags: List[str] = Field(default_factory=list)

    @property
    def _calc_name(self):
        if not self.name:
            if self.host:
                return f"{self.host.name}-{self.port}"
            return f"{self.host_name}-{self.port}"
        return self.name

    def __lt__(self, other):
        return self._calc_name < other._calc_name

    def __eq__(self, other):
        return self._calc_name == other._calc_name

    def __hash__(self):
        return hash(self._calc_name)

    @property
    def pool(self):
        return self._calc_name

    @model_validator(mode="after")
    def _validate(self) -> Self:
        if not self.host and not self.host_name:
            raise ValueError("Either host or host_name must be provided")
        if self.host and self.host_name and self.host.name != self.host_name:
            raise ValueError("Host and host_name must match")
        return self



from contextlib import contextmanager
from unittest.mock import patch

__all__ = ("pools", "variables")


@contextmanager
def pools(return_value=None, side_effect=None):
    with patch("airflow_balancer.config.balancer.Pool") as pool_mock:
        pool_mock.get_pool.return_value = return_value
        if side_effect:
            pool_mock.get_pool.side_effect = side_effect
        yield pool_mock


@contextmanager
def variables(return_value=None, side_effect=None):
    with patch("airflow.models.variable.Variable.get") as get_mock:
        get_mock.return_value = return_value
        if side_effect:
            get_mock.side_effect = side_effect
        yield get_mock



from typing import List, Optional

from airflow.models.variable import Variable
from airflow.providers.ssh.hooks.ssh import SSHHook
from pydantic import BaseModel, Field

__all__ = ("Host",)


class Host(BaseModel):
    name: str
    username: Optional[str] = None

    # Password
    password: Optional[str] = None
    # If password is stored in a variable
    password_variable: Optional[str] = None
    # if stored in structured container, access by key
    password_variable_key: Optional[str] = None
    # Or get key file
    key_file: Optional[str] = None

    os: Optional[str] = None

    # Airflow / balance
    pool: Optional[str] = None
    size: Optional[int] = None
    queues: List[str] = Field(default_factory=list)

    tags: List[str] = Field(default_factory=list)

    def override(self, **kwargs) -> "Host":
        return Host(**{**self.model_dump(), **kwargs})

    def hook(self, username: str = None, use_local: bool = True, **hook_kwargs) -> SSHHook:
        if use_local and not self.name.count(".") > 0:
            name = f"{self.name}.local"
        else:
            name = self.name
        username = username or self.username
        if username and self.password:
            return SSHHook(remote_host=name, username=username, password=self.password, **hook_kwargs)
        elif username and self.password_variable:
            if self.password_variable_key:
                credentials = Variable.get(self.password_variable, deserialize_json=True)
                password = credentials[self.password_variable_key]
            else:
                password = Variable.get(self.password_variable)
            return SSHHook(remote_host=name, username=username, password=password, **hook_kwargs)
        elif username and self.key_file:
            return SSHHook(remote_host=name, username=username, key_file=self.key_file, **hook_kwargs)
        elif username:
            return SSHHook(remote_host=name, username=username, **hook_kwargs)
        else:
            return SSHHook(remote_host=name, **hook_kwargs)

    def __lt__(self, other):
        return self.name < other.name

    def __le__(self, other):
        return self.name <= other.name

    def __hash__(self):
        return hash(self.name)
